/*	Timidity Wrapper Plugin for XMMS
 *	Copyright (C) 2003  Adam Nielsen
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

// TODO: If a bad Timidity option is given, XMMS won't close and will probably
//       lock up (because it's waiting for the FIFO to be opened, but Timidity
//       quit so it never will be opened.

// TODO: Change the buffering so "(titles)(like)(this)" can still be read if
//       there's no newline between them.

// TODO: Seperate Timidity's standard output - at the moment it's mixed in with
//       XMMS' output which isn't so good.  It means that anything XMMS or any
//       other plugins print is checked to see if there's a song-title in it...

// TODO: Implement seeking.  This is currently impossible with the 'dumb'
//       Timidity interface, so a new interface would have to be used.

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <errno.h>

#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <pthread.h>
#include <ctype.h>
#include <assert.h>

#include <xmms/plugin.h>
#include "xmms-timidity.h"

// Define this if you want Timidity's output repeated to XMMS' stdout.
//#define __TOUTPUT

typedef unsigned char byte;

byte *pSoundData = NULL;

char cTitle[256];
int iTitlePriority = 0;
int iSongLength = 0;
bool bPaused = false;

#define BUF_SIZE						SAMPLE_RATE         // Enough for a second
#define SAMPLE_SIZE					(BITS_PER_SAMPLE/8)

#if (BITS_PER_SAMPLE == 8)
	#define S_FMT FMT_S8
#else
	#define S_FMT FMT_S16_LE	// Default to 16-bit
#endif

#define SAMPLES_PER_BLOCK		512

void readSZ(FILE *hFile, char *cFile, int iMaxLen);

bool bPlaying = false;

pthread_t playThreadID = 0, infoThreadID = 0;
char cPipeName[80], cConfigName[80];
int tpipe;

void pluginInit(void);
void pluginAbout(void);
void pluginConfig(void);
int pluginCanPlay(char *cFilename);
void pluginPlay(char *cFilename);
void *infoThread(void *pData);
void *playThread(void *pData);
//void getSongTitle(FILE *hFile, char *cFilename, char *cTitle, int iMaxLen);
void pluginStop(void);
void pluginPause(short iPaused);
void pluginSeek(int iTime);
int pluginGetTime(void);
void pluginClose(void);
void pluginGetSongInfo(char *cFilename, char **cTitle, int *iLength);
void pluginInfo(char *cFilename);
//CPlayFile *findInstance(char *cFilename);

InputPlugin ipTimidity = {
	NULL,									// handle, XMMS fills in
	NULL,									// filename, XMMS fills in
	"Timidity Wrapper",		// description
	pluginInit,						// void init(void), called when plugin is loaded
	pluginAbout,					// void about(void), called to show about box
	pluginConfig,					// void configure(void), called to show config box
	pluginCanPlay,				// int is_our_file(char *filename), returns 1 if we can handle the file
	NULL,									// GList *scan_dir(char *dirname), no idea (see Input/cdaudio/cdaudio.c)
	pluginPlay,						// void play_file(char *filename), start playback
	pluginStop,						// void stop(void), stop playback
	pluginPause,					// void pause(short paused), pause/resume
	pluginSeek,						// void seek(int time), seek to the specified time (in seconds?)
	NULL,									// void set_eq(int on, float preamp, float *bands), set the equalizer, most plugins won't be able to do this
	pluginGetTime,				// int get_time(void), usually returns the output plugin's output time
	NULL,									// void get_volume(int *l, int *r), input-specific volume functions, NULL means output plugin will handle it
	NULL,									// void set_volume(int l, int r), ditto
	pluginClose,					// void cleanup(void), called when xmms exits
	NULL,									// get_vis_type is obsolete
	NULL,									// void add_vis_pcm(int time, AFormat fmt, int nch, int length, void *ptr), sends data to the vis plugins (pref 512 samples/block)
	NULL,									// void set_info(char *title, int length, int rate, int freq, int nch), fill in info for player window.  length == -1 if unknown.  Filled in by XMMS.
	NULL,									// void set_info_text(char *text), show text in the song title box, set to NULL to reset it to the song title.  Filled in by XMMS.
	pluginGetSongInfo,		// void get_song_info(char *filename, char **title, int *length), get title string
	pluginInfo,						// void file_info_box(char *filename), show file info box for supplied file
	NULL									// OutputPlugin *output, points to current output plugin, filled in by XMMS.
};

//CPlayFile *pPlayer = NULL;

extern "C" InputPlugin *get_iplugin_info(void)
{
	return &ipTimidity;
}

void pluginInit(void)
{
	// Create the FIFO
	char *cUserDir = getenv("HOME");
	// TODO: Make sure strlen(HOME) < sizeof(cPipeName) - strlen(PIPE_NAME) - 1
	strcpy(cPipeName, cUserDir);
	strcat(cPipeName, "/");
	strcat(cPipeName, PIPE_NAME);//"-XXXXXX");
	printf("xmms-timidity: Creating pipe in %s\n", cPipeName);
	//mktemp(cPipeName);
	//printf("Creating fifo '%s'\n", cPipeName);
//	unlink(cPipeName); // Try to delete first
	if (mkfifo(cPipeName, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP)) {
		if (errno != EEXIST) {
			char *cMsg = new char[50+strlen(cPipeName)];
	    sprintf(cMsg, "Unable to create audio pipe \"%s\"", cPipeName);
			perror(cMsg);
			delete cMsg;
		}
	}
	// Create the temp extra config
	// TODO: Make sure strlen(HOME) < sizeof(cPipeName) - strlen(PIPE_NAME) - 1
	strcpy(cConfigName, cUserDir);
	strcat(cConfigName, "/");
	strcat(cConfigName, CONFIG_NAME);//"-XXXXXX");
	printf("xmms-timidity: Creating temporary Timidity config in %s\n", cConfigName);
	//mktemp(cConfigName);
	//printf("Creating extra config '%s'\n", cConfigName);
	return;
}

void pluginAbout(void)
{
	return;
}

void pluginConfig(void)
{
	printf("Config not implemented.\n");
	return;
}

int pluginCanPlay(char *cFilename)
{
	int iExtPos = strlen(cFilename) - 3;	// only three letter extensions needed
	char cExt[4];
	for (int i=0;i<3;i++) {
		cExt[i] = tolower(cFilename[iExtPos + i]);
	}
	cExt[3] = 0;
	// extension will always be lowercase now
	if (strcmp(cExt, "mid") == 0) return 1;
	if (strcmp(cExt, "rmi") == 0) return 1;
	return 0;
}


void pluginPlay(char *cFilename)
{
	// Change into the directory containing the file first, because otherwise
	// Timidity won't be able to load any SoundFont if the path has a space in it
	int iSize = strlen(cFilename) + 1;
	char *cPath = new char[iSize];
	char *cFile = new char[iSize];
	memcpy(cPath, cFilename, iSize);
	cFile[0] = 0;
	for (int i = iSize; i > 0; i--) {
		if (cPath[i] == '/') {
			cPath[i] = 0;
			strcpy(cFile, &cPath[i+1]);
			break;
		}
	}
	if (cFile[0] == 0) {
		// there was no path, just a filename
		printf("Warning: '%s' does not contain a path.\n", cFilename);
		cPath[0] = 0;
		strcpy(cFile, cFilename);
	} else chdir(cPath);
	
	iSize = strlen(cFile) + 1;
	// Check if there's a SoundFont in the current directory and if so add it to
	// the extra config file.
	char *cSF = new char[iSize];
	memcpy(cSF, cFile, iSize);
	strcpy(&cSF[iSize-4], "SF2");
	struct stat buf;
	bool bSoundFont = true;
	if (stat(cSF, &buf) != 0) {
		// No SF2, try sf2
		strcpy(&cSF[iSize-4], "sf2");
		if (stat(cSF, &buf) != 0) {
			// No sf2, try Sf2
			strcpy(&cSF[iSize-4], "Sf2");
			if (stat(cSF, &buf) != 0) {
				// No Sf2, try sF2
				strcpy(&cSF[iSize-4], "sF2");
				if (stat(cSF, &buf) != 0) {
					// No sF2, skip config
					bSoundFont = false;
				} // sF2
			} // Sf2
		} // sf2
	} // SF2
	if (bSoundFont) {
		// There is a soundfont
		printf("Found SoundFont, telling Timidity to load %s.\n", cSF);
		FILE *hConfig = fopen(cConfigName, "wt");
		fprintf(hConfig, "soundfont %s\n", cSF);
		fclose(hConfig);
	}

	int pipeTimidity[2];
	pipe(pipeTimidity);
	// TODO: error

	pid_t forkResult = fork();
	switch (forkResult) {
		case -1:
			fprintf(stderr, "Error during fork.  This shouldn't happen unless something is really in trouble...\n");
			return;
		case 0:
			// Child
			
			// Generate the option for the output sample rate
			char cSampleRate[30];
			sprintf(cSampleRate, "-s%u", SAMPLE_RATE);

			// Generate the option for the output format
			char cFormat[30];
			strcpy(cFormat, "-Or");
			if (NUM_CHANNELS == 2) strcat(cFormat, "S");	// Stereo
			else strcat(cFormat, "M");	// Mono
			if (SAMPLE_SIZE == 2) strcat(cFormat, "1");	// 16-bit
			else strcat(cFormat, "8");	// 8-bit

			if (bSoundFont) {
				printf("xmms-timidity: exec'ing %s %s %s %s %s %s %s %s %s %s\n", TIMIDITY_EXE, cFile, "-EFreverb=0", cFormat, "-o", ::cPipeName, "-idv", cSampleRate, "-c", cConfigName);
			} else {
				printf("xmms-timidity: exec'ing %s %s %s %s %s %s %s %s\n", TIMIDITY_EXE, cFile, "-EFreverb=0", cFormat, "-o", ::cPipeName, "-idv", cSampleRate);
			}
			//int iNewOut = dup(STDOUT_FILENO);
			//setvbuf(iNewOut, NULL, _IONBF, 0);
			//fdopen(STDOUT_FILENO, "r");
			setvbuf(stdout, NULL, _IONBF, 0);
			dup2(pipeTimidity[1], STDOUT_FILENO);	// Data going into pipe is now our stdout
			close(pipeTimidity[0]);	// Close read-side of pipe
			close(pipeTimidity[1]);	// Close the original copy of the write-side of the pipe
			setvbuf(stdout, NULL, _IONBF, 0);
			
			if (bSoundFont) {
				execlp(TIMIDITY_EXE, TIMIDITY_EXE, cFile, "-EFreverb=0", cFormat, "-o", ::cPipeName, "-idv", cSampleRate, "-c", cConfigName, NULL);
			} else {
				execlp(TIMIDITY_EXE, TIMIDITY_EXE, cFile, "-EFreverb=0", "-Or", "-o", ::cPipeName, "-idv", cSampleRate, NULL);
			}
			// Error executing Timidity, open/close the pipe for writing so as not to
			// lock up the other thread when it goes to read from it.
			::bPlaying = false;
			{
				// TODO: Doesn't seem to have the intended affect at the moment, all it
				// seems to do is make XMMS crash...
				FILE *hPipe = fopen(::cPipeName, "wb");
				//fwrite("error\n", 6, 1, hPipe);
				fclose(hPipe);
			}
			return;
		default:
			// Parent
			close(pipeTimidity[1]);	// Close write-side of pipe
			::tpipe = pipeTimidity[0];
			// forkResult now contains PID of child
			break;
	}

	// int iChild;
	// close(pipeTimidity[0]);	// EOF for child
	// wait(&iChild);	// Wait for any child processes to exit

	if (playThreadID != 0) {
		//printf("already playing, stop (bad)");
		// already playing, stop
		::ipTimidity.output->pause(0);	// unpause first, just in case
		// TODO: can you check if the output's paused first?
		pluginStop();
	} //else printf("not playing atm (good)");

	::bPlaying = true;

	iTitlePriority = 0;
	iSongLength = 0;

	int iFilenameLen = strlen(cFilename);
	int i = iFilenameLen - 1;	// filename will always be at least one character, and this will avoid a trailing slash
	while (cFilename[i] != '/') i--;
	iFilenameLen -= i;	// Includes terminating NULL (removed later)
	//this->cFilename = new char[this->iFilenameLen];
	strncpy(cTitle, &cFilename[i+1], iFilenameLen);

	pthread_create(&playThreadID, NULL, playThread, (void *)forkResult);
	pthread_create(&infoThreadID, NULL, infoThread, (void *)pipeTimidity[0]);

	return;
}

void setSongTitle(void)
{
	int iLen = strlen(::cTitle);
	char *cTitle = (char *)g_malloc(iLen + 1);
	memcpy(cTitle, ::cTitle, iLen + 1);
	::ipTimidity.set_info(cTitle, ::iSongLength, (SAMPLE_RATE * NUM_CHANNELS * SAMPLE_SIZE) * 8192, SAMPLE_RATE, NUM_CHANNELS);
	return;
}

void *infoThread(void *pData)
{
	int pipeTimidity = (int)pData;
	char cLine[256];
	signed int iBytes = 1;
	while ((iBytes > 0) && (::bPlaying)) {
		bool bLine = false;
		for (int i=0;i<256;i++) {
			iBytes = read(pipeTimidity, &cLine[i], 1);
			if (iBytes == -1) break;	// EOF (usually means we've stopped playback and need to exit thread)
			if (cLine[i] == '\n') {	// EOL
				cLine[i] = 0;
				bLine = true;
				break;
			}
		}
		if (!bLine) {
			// Line is too long
			char cChar = 0;
			while ((cChar != '\n') && (iBytes > 0) && (::bPlaying)) {
				iBytes = read(pipeTimidity, &cChar, 1);
			}
		}
		//
		// Songs to test this with: Everybody Off.mid doesn't have a trailing dot, but all the other Shadow Warrior ones do
		//                          Blood_I.mid cuts of Daniel Bernstein's name
		//
#ifdef __TOUTPUT
		fprintf(stderr, "Timidity output: '%s'\n", cLine);
#endif
		if (::iTitlePriority < 3) {
			if (strncmp(cLine, "Sequence: ", 10) == 0) {
				// title starts at cLine[10]
				if (strncasecmp(&cLine[10], "untitled", 256-10) != 0) {
					strncpy(::cTitle, &cLine[10], 256-10);
					cTitle[255] = 0;
					setSongTitle();
					::iTitlePriority = 3;	// 'Sequence' is the highest priority for a title
				} // else title is "untitled" so try the next one
			} else if (::iTitlePriority < 2) {
				if (strncmp(cLine, "Text: ", 6) == 0) {
					// title starts at cLine[6]
					strncpy(::cTitle, &cLine[6], 256-6);
					cTitle[255] = 0;
					cTitle[strlen(cTitle)] = 0;	// hide trailing dot that Timidity adds
					setSongTitle();
					::iTitlePriority = 2;	// 'Sequence' is the highest priority for a title
				} else if (::iTitlePriority < 1) {
					if (cLine[0] == '(') {
						// TODO: This does not work because there aren't any newlines after
						// the 'markers' in brackets that we check for here

						// Title starts at cLine[1] and continues until the next close
						// bracket (unless there are any opening brackets, in which case
						// allow one close bracket through for every open one)
						int iNumBrackets = 0;
						for (int i=1;i<256;i++) {
							if (cLine[i] == '(') iNumBrackets++;
							if (cLine[i] == ')') iNumBrackets--;
							if (iNumBrackets < 0) {
								cLine[i] = 0;
								break;
							}
						}
						strncpy(::cTitle, &cLine[1], 256-1);
						cTitle[255] = 0;
						setSongTitle();
						::iTitlePriority = 1;	// 'Sequence' is the highest priority for a title
					}
				}
			} // else priority == 2 so we've already got a title *that* way
		}
		if (::iSongLength == 0) {
			// Haven't found time yet, so keep checking
			char *cStrStart = strstr(cLine, "supported events,");
			if (cStrStart != NULL) {
				cStrStart += 18; // skip to just after "supported events, "
				::iSongLength = (int)(((double)atoi(cStrStart)) / (((double)SAMPLE_RATE)/1000.0));
				setSongTitle();
			}
		}
	}
	close(pipeTimidity);
	pthread_exit(NULL);
}

void *playThread(void *pData)
{
//	pid_t pidTimidity = (pid_t)pData;
	::ipTimidity.output->close_audio();
	::ipTimidity.output->open_audio(S_FMT, SAMPLE_RATE, NUM_CHANNELS);
	//::ipTimidity.output->flush(0);

	unsigned int pBlock[SAMPLES_PER_BLOCK * SAMPLE_SIZE * NUM_CHANNELS + 1];
	unsigned long iTime = 0;
	
	// TODO: Test the pipe first and wait until Timidity has opened it, or
	// timeout or something so that XMMS doesn't lock up if Timidity does run
	// as expected.
	FILE *hPipe = fopen(::cPipeName, "rb");

	size_t iNumSamples = SAMPLES_PER_BLOCK;// * SAMPLE_SIZE;
	//while ((::bPlaying) && (iNumSamples >= (SAMPLES_PER_BLOCK * SAMPLE_SIZE))) {
	while ((::bPlaying) && (iNumSamples >= SAMPLES_PER_BLOCK)) {
		iNumSamples = fread(pBlock, NUM_CHANNELS * SAMPLE_SIZE, SAMPLES_PER_BLOCK, hPipe);
		//printf("iNumSamples == %d/%d\n", iNumSamples, SAMPLES_PER_BLOCK);

		while ((unsigned)(::ipTimidity.output->buffer_free())  <= (iNumSamples * NUM_CHANNELS * SAMPLE_SIZE)) {
			usleep(10000);
			// TODO: check for early awake signals (maybe, it's pretty short, might not need it
			// TODO: usleep() obsolete, use nanosleep() or something instead
		}

		::ipTimidity.output->write_audio(pBlock, iNumSamples * NUM_CHANNELS * SAMPLE_SIZE);
		// TODO: Change 512.0 to (double)SAMPLES_PER_BLOCK?
		::ipTimidity.add_vis_pcm((int)((iTime*512.0) / (((double)SAMPLE_RATE)/1000.0)), S_FMT, NUM_CHANNELS, iNumSamples * SAMPLE_SIZE, pBlock);
		//printf("time is %lu, ms == %lu\n", iTime, (iTime*512) / FREQ_HZ);
		iTime++;//= iNumSamples;
	}

	//kill timidity?  yes, because we might be here even if we haven't reached the end (i.e. "stop" was pressed)
	//kill(pidTimidity, SIGTERM);
	// Perhaps just closing the pipe would be enough
	fclose(hPipe);
//	int iStatus = 0;
	//printf("waiting for timidity to close\n");
	//waitpid(pidTimidity, &iStatus, 0);
	//printf("timidity closed\n");
	while ((::bPlaying) && (::ipTimidity.output->buffer_playing())) usleep(10000);

	::bPlaying = false;

	::ipTimidity.output->flush(0);
	::ipTimidity.output->close_audio();
	pthread_exit(NULL);
}

void pluginStop(void)
{
	::bPlaying = false;
	void *pThreadReturn;
	if (playThreadID != 0) {
		if (::bPaused) ::ipTimidity.output->pause(0);	// unpause first, just in case
		pthread_join(playThreadID, &pThreadReturn);
		playThreadID = 0;
	}
	if (infoThreadID != 0) {
		close(::tpipe);
		pthread_join(infoThreadID, &pThreadReturn);
		infoThreadID = 0;
	}
	return;
}

void pluginPause(short iPaused)
{
	::ipTimidity.output->pause(iPaused);
	::bPaused = (iPaused != 0);
	return;
}

void pluginSeek(int iTime)
{
	return;
}

int pluginGetTime(void)
{
	if (!::bPlaying) return -1;
	return ::ipTimidity.output->output_time();
}

void pluginClose(void)
{
	if (::bPlaying) {
		printf("ERROR: pluginClose() called when still outputting audio.");
		pluginStop();
	}
	unlink(::cPipeName);
	unlink(::cConfigName);
	return;
}

void pluginGetSongInfo(char *cFilename, char **cTitle, int *iLength)
{
	if (cTitle != NULL) { //&& (*cTitle != NULL)) {
		*cTitle = (char *)g_malloc(256);

		int iFilenameLen = strlen(cFilename);
		int i = iFilenameLen - 1;	// filename will always be at least one character, and this will avoid a trailing slash
		while (cFilename[i] != '/') i--;
		iFilenameLen -= i;	// Includes terminating NULL (removed later)
		*cTitle = (char *)g_malloc(iFilenameLen);
		strncpy(*cTitle, &cFilename[i+1], iFilenameLen);
	}
	if (iLength != NULL) {
		*iLength = -1; // Unknown length
	}
	return;
}

void pluginInfo(char *cFilename)
{
	return;
}
